import sys
import os
import unittest
import tempfile

sys.path.append("..")

from kmerdb import database

class Test_parsefile(unittest.TestCase):

    def setUp(self):

        self.k = 3
        self.database  = tempfile.NamedTemporaryFile(mode="w+", suffix=".sqlite3", delete=False)
        self.database.close()


    def test_str_filepath_only(self):
        """
        SqliteKdb throws a TypeError if it receives anything but a str
        """
        with self.assertRaises(TypeError):
            database.SqliteKdb(None, self.k)
        with self.assertRaises(TypeError):
            database.SqliteKdb(1, self.k)
        with self.assertRaises(TypeError):
            database.SqliteKdb(1.0, self.k)
        with self.assertRaises(TypeError):
            database.SqliteKdb([1], self.k)
        with self.assertRaises(TypeError):
            database.SqliteKdb({'hello': 'world'}, self.k)

    def test_filepath_exists(self):
        """
        SqliteKdb throws an OSError if it receives a filepath that doesnt exist
        """
        with self.assertRaises(OSError):
            db = database.SqliteKdb("helloworld.txt", self.k)
            db.conn.close()
            os.unlink("helloworld.txt")

    def test_k_is_int_only(self):
        """
        SqliteKdb throws a TypeError if it receives anything but an int
        """
        with self.assertRaises(TypeError):
            database.SqliteKdb(self.database, None)
        with self.assertRaises(TypeError):
            database.SqliteKdb(self.database, True)
        with self.assertRaises(TypeError):
            database.SqliteKdb(self.database, 'hello')
        with self.assertRaises(TypeError):
            database.SqliteKdb(self.database, 1.0)
        with self.assertRaises(TypeError):
            database.SqliteKdb(self.database, [1])
        with self.assertRaises(TypeError):
            database.SqliteKdb(self.database, {'hello': 'world'})


    def test_returns_database(self):
        """
        SqliteKdb returns a database object
        """

        testDb = database.SqliteKdb(self.database.name, self.k)
        self.assertTrue(isinstance(testDb, database.SqliteKdb))


if __name__ == '__main__':
    unittest.main()



import sys
import os
import unittest

sys.path.append("..")

from kmerdb import parse, database

class Test_parsefile(unittest.TestCase):

    def setUp(self):

        self.k = 3
        self.fasta  = "test/data/Ctetani_E88.fasta.gz"


    def test_str_filepath_only(self):
        """
        kmer_to_id throws a TypeError if it receives anything but a str
        """
        with self.assertRaises(TypeError):
            parse.parsefile(None, self.k)
        with self.assertRaises(TypeError):
            parse.parsefile(1, self.k)
        with self.assertRaises(TypeError):
            parse.parsefile(1.0, self.k)
        with self.assertRaises(TypeError):
            parse.parsefile([1], self.k)
        with self.assertRaises(TypeError):
            parse.parsefile({'hello': 'world'}, self.k)

    def test_filepath_exists(self):
        """
        parsefile throws an OSError if it receives a filepath that doesnt exist
        """
        with self.assertRaises(OSError):
            db, header = parse.parsefile("foobar.txt", self.k)
            db.conn.close()
            os.unlink("foobar.txt")

    def test_k_is_int_only(self):
        """
        parsefile throws a TypeError if it receives anything but an int
        """
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, None)
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, True)
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, 'hello')
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, 1.0)
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, [1])
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, {'hello': 'world'})
        
    def test_p_is_int_only(self):
        """
        parsefile throws a TypeError if it receives anything but an int
        """
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, p=None)
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, p=True)
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, p='hello')
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, p=1.0)
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, p=[1])
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, p={'hello': 'world'})

    def test_b_is_int_only(self):
        """
        parsefile throws a TypeError if it receives anything but an int
        """
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, b=None)
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, b=True)
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, b='hello')
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, b=1.0)
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, b=[1])
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, b={'hello': 'world'})

    def test_n_is_int_only(self):
        """
        parsefile throws a TypeError if it receives anything but an int
        """
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, n=None)
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, n=True)
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, n='hello')
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, n=1.0)
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, n=[1])
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, n={'hello': 'world'})

            
    def test_stranded_is_bool_only(self):
        """
        parsefile throws a TypeError if it receives anything but an int
        """
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, stranded=None)
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, stranded='hello')
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, stranded=1)
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, stranded=1.0)
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, stranded=[1])
        with self.assertRaises(TypeError):
            parse.parsefile(self.fasta, self.k, stranded={'hello': 'world'})


    def test_returns_database_and_header_dictionary(self):
        """
        parsefile returns a header dictionary object
        """
        testdb, headerDict, n = parse.parsefile(self.fasta, self.k)
        self.assertDictEqual({'filename': 'test/data/Ctetani_E88.fasta.gz',
                              'md5': '2d6e56382e6b296cd7aa00d33d6a3a25',
                              'mononucleotides': {'A': 1016563, 'C': 418556, 'G': 404380, 'T': 1033834},
                              'nullomers': 0,
                              'sha256': '61b5cf99700219774f05beb9203f0388d95273b227ae7ae37164f1c9a53665ca',
                              'total_kmers': 2873329,
                              'total_reads': 2,
                              'unique_kmers': 64}
                             , headerDict)
        self.assertTrue(isinstance(testdb, database.SqliteKdb))        
            

if __name__ == '__main__':
    unittest.main()



'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''




import logging
logger = logging.getLogger(__file__)


import io
import os
import json
#import sqlite3
import sys
import array
from sqlalchemy import create_engine
from sqlalchemy import Table, Column, Integer, String, MetaData, ForeignKey, Sequence, Text, Boolean
from sqlalchemy.pool import NullPool
#from sqlalchemy.ext.declarative import declarative_base
#from sqlalchemy.orm import sessionmaker

rows_per_loading_transaction = 1000000
#Base = declarative_base()


# ins = kmers.insert()
# conn = engine.connect()
# conn.execute(ins, [{'count': 0}, ...])


def histogram(conn):
    res = conn.execute("SELECT COUNT(count) FROM kmers GROUP BY count ORDER BY count")
    hist = array.array('H')
    for c in res.fetchall():
        hist.append(c[0])
    return hist


class SqliteKdb:
    def __init__(self, filename:str, k:int):
        
        if type(filename) is not str:
            raise TypeError("kmerdb.database.SqliteKdb expects a str as its first positional argument")
        elif not os.path.exists(filename):
            raise OSError("kmerdb.database.SqliteKdb expects an existing .kmerdb file as its first positional argument")
        elif type(k) is not int:
            raise TypeError("kmerdb.database.SqliteKdb expects an int as its second positional argument")
        self._max_records = 4**k
        self.filepath = filename
        self._engine = create_engine("sqlite:///{0}".format(filename), echo=False, poolclass=NullPool)
        if not os.path.exists(filename) or os.path.getsize(filename) == 0:
            logger.info("No data exists in '{0}', initializing empty database".format(filename))
            ##SQLAlchemy 1.3
            self.__make_empty_database()
        self.conn = self._engine.connect()
        self.__validate_database()
        self._id = 1

        

        
    def __make_empty_database(self):
        metadata = MetaData()
        kmers = Table('kmers', metadata,
                      Column('id', Integer, Sequence('kmer_id_seq'), primary_key=True),
                      Column('count', Integer),
                      Column('starts', Text),
                      Column('reverses', Text),
                      Column('seqids', Text)
        )
        # reads = Table('reads', metadata,
        #               Column('read_id', String),
        #               Column('kmer_id', None, ForeignKey('kmers.id'))
        # )
        # Create tables
        metadata.create_all(self._engine)
        #null_profile = array.array('B')
        null_profile = []
        # FIXME This won't work for certain values of k
        for x in range(self._max_records):
            null_profile.append({
                'count': 0,
                'starts': '[]',
                'reverses': '[]',
                'seqids': '[]'
            })

        with self._engine.connect() as conn:
            while len(null_profile) > 0:
                temp = []
                for y in range(rows_per_loading_transaction):
                    if len(null_profile) > 0:
                        temp.append( null_profile.pop(0) )
                conn.execute(kmers.insert(), temp)

            
    def __exit__(self, exc_type, exc_value, traceback):
        self.conn.close()
        self._engine.dispose()

    def __validate_database(self):
        assert self._get_total_records() == self._max_records, "The number of total records in the database should equal 4^k"

    def _get_total_records(self):
        res = self.conn.execute("SELECT COUNT(id) FROM kmers")
        return res.fetchone()[0]
        
    def _get_nullomers(self):
        res = self.conn.execute("SELECT COUNT(id) FROM kmers WHERE count = 0")
        return res.fetchone()[0]

    def _get_sum_counts(self):
        res = self.conn.execute("SELECT SUM(count) from kmers")
        return res.fetchone()[0]

    def _load_metadata(self):
        """ Not implemented """
        return

    
    def __iter__(self):
        return self

    def __next__(self):
        res = self.conn.execute("SELECT * FROM kmers WHERE id = ?", self._id)
        self._id += 1
        results = res.fetchone()
        if results is None or len(results) == 0:
            raise StopIteration
        else:
            return results


    
# class Kmer(Base):
#     __tablename__ = 'kmers'

#     id = Column(Integer, Sequence('kmer_id_seq'), primary_key=True)
#     count = Column(Integer)


#     def __repr__(self):
#         return "<Kmer(id='%s', count='%s')>" % (self.id, self.count)


    



'''
   Copyright 2020 Matthew Ralston

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

'''



VERSION="0.0.7"
end_header_line = "    kdb: can someone get me some sentries"
header_delimiter = end_header_line + "\n" + ("="*24) + "\n"

metadata_schema = {
    "type": "object",
    "properties": {
        "version": {"type": "string"},
        "metadata_blocks": {"type": "number"},
        "k": {"type": "number"},
        "total_kmers": {"type": "number"},
        "unique_kmers": {"type": "number"},
        "metadata": {"type": "boolean"},
        "tags": {
            "type": "array",
            "items": {"type": "string"}
        },
        "files": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "filename": {"type": "string"},
                    "sha256": {
                        "type": "string",
                        "minLength": 64,
                        "maxLength": 64
                    },
                    "md5": {
                        "type": "string",
                        "minLength": 32,
                        "maxLength": 32
                    },
                    "total_reads": {"type": "number"},
                    "total_kmers": {"type": "number"},
                    "unique_kmers": {"type": "number"},
                    "nullomers": {"type": "number"},
                    "mononucleotides": {
                        "type": "object",
                        "properties": {
                            "A": {"type": "number"},
                            "C": {"type": "number"},
                            "G": {"type": "number"},
                            "T": {"type": "number"}
                        }
                    }
                },
                "required": ["filename", "sha256", "md5", "total_reads", "total_kmers", "unique_kmers", "nullomers"]
            }
        },
        "comments": {
            "type": "array",
            "items": {"type": "string"}
        }
    },
    "required": ["version", "metadata_blocks", "k", "tags", "files"]
}




























pca_variance_fig_filepath = "PCA_variance_accumulation.png"
kmeans_elbow_graph_fig_filepath = "kmeans_elbow_graph.png"
kmeans_clustering_fig_filepath = "kmeans_clustering_of_kmer_profiles.png"
ecopy_rarefaction_fig_filepath = "ecopy_rarefaction_curve.png"
hierarchical_clustering_dendrogram_fig_filepath = "dendrogram.png"
spearman_upgma_tree_phy = "kdb_spearman_upgma_tree.phyloxml"
files = (pca_variance_fig_filepath, kmeans_elbow_graph_fig_filepath, kmeans_clustering_fig_filepath, ecopy_rarefaction_fig_filepath, hierarchical_clustering_dendrogram_fig_filepath)

DEFAULT_MASTHEAD = """

==========================================
               k m e r d b
==========================================

Thank you for using kmerdb. Please feel free
to submit issues and requests to 
https://github.com/MatthewRalston/kdb

Copyright 2020 Matt Ralston (mrals89@gmail.com)

# First links
https://matthewralston.github.io/blog/kmer-database-format-part-1
https://github.com/MatthewRalston/kdb


Please cite my repository in your work!

Feel free to e-mail me or reach out!

"""
DONE = """

==========================================
----------------D O N E-------------------
==========================================

"""


DISTANCE_MASTHEAD = """

==========================================
           kmerdb distance
==========================================

Distance matrix generation beginning!

Distance matrix will be written to STDOUT as this is the first step of the pipeline.

"""
for i in range(42, 1, -1):
    DISTANCE_MASTHEAD += i*"=" + "\n"


MATRIX_MASTHEAD = """

==========================================
           kmerdb matrix
==========================================

Matrix generation beginning!

Matrix will be written to STDOUT as this is the first step of the pipeline.

"""
for i in range(42, 1, -1):
    MATRIX_MASTHEAD += i*"=" + "\n"


KMEANS_MASTHEAD = """

==========================================
           kmerdb kmeans
==========================================

K-means clustering beginning!

"""
for i in range(42, 1, -1):
    KMEANS_MASTHEAD += i*"=" + "\n"


HIERARCHICAL_MASTHEAD = """

==========================================
           kmerdb hierarchical
==========================================

Hierarchical clustering beginning!

"""
    


DEBUG_MASTHEAD = '''

The workflow is roughly as follows:

# # # # # # # # # #  #
# profile generation #
# # # # # # # # # #  #
# I have included -p and -b parameters that influence the rate of profile generation.
# The block size (-b) is primarily for the number of .fastq(.gz) records to read at a time.
# The -p parallel feature works within fastq parsing to process k-mer shredding/counting.
#
# -k $K is the parameter that specifies the k-mer resolution
#
# This command uses SQLite3 behind the scenes for on-disk k-mer counting
# since memory is rate limiting for profile generation when dealing 
# with biologically conventional choices of k (20 < k < 35).
parallel 'kmerdb profile -k $K {{}} {{.}}.$K.kdb' ::: $(/bin/ls test/data/*.fasta.gz)



# # # # # # #
# analysis  #
# # # # # # #
################################
# W A R N I N G :  M E M O R Y #
################################
# The first step of either rarefaction or clustering is to generate the k-mer profile matrix
# The matrix is not a new concept, it is just samples as columns of profiles. 
# Without metadata or graphical information embedded in the format, 
# we can create a matrix and do data science with the information.
# However, because the profiles are exponential in k, a linear increase in k 
# hypothetically results in an improvement of resolution that is at least superlinear in k.
# Therefore, the amount of memory can be calculated by the integer size times the profile resolution
# times the number of samples.
#
#
##################
# PCA + k-means
##################
# The first step ('kdb matrix') generates one from different profiles with the same choice of k.
# This command uses ecopy to normalize between sample k-mer total counts before PCA/tSNE.
# -n $N is the dimensionality of either PCA or tSNE. A good choice for tSNE is 2.
# If the command is run with PCA without a selected dimensionality, an elbow graph
# will be produced named '{0}'. Please use this graph to select
# the number of principal components to use.
# The pipeline will not continue until -n $N is selected by the user.
# It is not recommended to feed Unnormalized or Normalized matrices directly to 'kdb kmeans'
# 
# The PCA/tSNE matrix will be dimReduced ($N) * N, where N is the number of samples/files/profiles.
#
# And finally, a k-means clustering will be done on the reduced dimensionality dataset
# Please note the randomness parameter 'random_state=42' for sklearn's kmeans is fixed at 42.
# Note here that the -k $K is not related to the choice of substring length 'k' for profile generation.
# The 'kdb kmeans' command produces two figures, first is an elbow graph looking at up to N clusters.
# This elbow graph will be written to '{1}'.
# The second is the more typical scatterplot of the first two reduced dimensions
# and the k-means clustering labels shown over the scatter.
# This file will be written to '{2}'.
kmerdb matrix [-n $N] [ PCA | tSNE ] test/data/*.$K.kdb | kmerdb kmeans -k $K sklearn
kmerdb matrix [-n $N] [ PCA | tSNE ] test/data/*.$K.kdb | kmerdb kmeans -k $K --distance e Biopython
#
# If you wanted to save a matrix from kdb matrix for use on your own
# it is recommended that you consider gzip compressing it if it is the Normalized or Unnormalized matrix
# which we will see is used downstream in the rarefaction and hierarchical analytics pathways.
#

##################
# Hierarchical
##################
#
# The Normalized matrix goes to the distance subcommand, which can use any of scipy's pdist distances
# to form the m x m distance matrix.
# The third step (kdb hierarchical)  is to build a dendrogram with scipy.cluster.hierarchy.
# This final step produces a plot in addition to the tsvs produced in the prior steps,
# which can be captured as independent steps or with tee in a pipeline.
kmerdb matrix [ Normalized ] test/data/*.$K.kdb | kmerdb distance spearman | kmerdb hiearchical

'''.format(*files)

